#
# Copyright (c) Daniel Sheffield 2021 - 2022
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
from decimal import Decimal
from typing import Any, Callable, Iterable, List, Union

import urwid
from additional_urwid_widgets import IndicativeListBox
from urwid import numedit
from urwid.wimp import PopUpLauncher


class AutoCompleteEdit(urwid.Edit):
    signals = [ *urwid.Edit.signals, 'apply', 'open' ]

    def __init__(self, name: str, *args: Iterable, **kwargs):
        if isinstance(name, tuple):
            pallete, title = name
            self.name = title
            title = title.title()
            passthrough = (pallete, u'  ' if title.lower() == 'unit' else u'')
        else:
            self.name = name
            title = name.title()
            passthrough = u'  ' if name.lower() == 'unit' else u''
        
        super().__init__(passthrough, *args, **kwargs)
    
    def keypress(self, size, key):
        if key == 'enter':
            self._emit('apply', self.name)
            return
        elif key == 'delete':
            self.set_edit_text('')
            return
        
        return super().keypress(size, key)

class AutoCompleteFloatEdit(numedit.FloatEdit):
    signals = [ *urwid.Edit.signals, 'apply', 'open' ]

    def __init__(self, name: str, *args: Iterable, **kwargs: dict):
        self.last_val = None
        self.op = '='
        self.pallete = None
        if isinstance(name, tuple):
            self.pallete, self.name = name
            title = self.name.title()
            passthrough = (self.pallete, f'{self.op} ')
        else:
            self.name = name
            title = name.title()
            passthrough = title
        
        super().__init__(passthrough, *args, **kwargs)
    
    def update_caption(self):
        if self.pallete is not None:
            self.set_caption((self.pallete, f'{self.op} '))
        else:
            self.set_caption(f'{self.op} ')

    def set_op(self, op):
        self.op = op
        self.last_val = self.value()
        self.set_edit_text('')
        self.update_caption()
        
    def calc(self):
        x = self.last_val
        op = self.op
        if op in ('+', '-',):
            y = self.value() or Decimal(0.0)
            if op == '+':
                z = x + y
            else:
                z = x - y
        elif op in ('*', '/'):
            y = self.value() or Decimal(1.0)
            if op == '*':
                z = x * y
            else:
                z = x / y
        else:
            y = self.value() or Decimal(0.0)
            z = y
        
        self.op = '='
        self.update_caption()
        self.set_edit_text(f'{z:.2f}')

    def keypress(self, size, key):
        if isinstance(key, tuple):
            return
        ops = ('+', '-', '*', '/',)

        if key in ops:
            if self.op in ops:
                self.calc()
            self.set_op(key)
            return
        elif key == 'enter':
            if self.get_edit_text() == '' or self.value() == Decimal(0.0):
                self._emit('apply', self.name)
                return
            self.calc()
            return
        elif key == '=':
            self.calc()
            return
        elif key == 'delete':
            self.set_edit_text('')
            self.op = '='
            self.update_caption()
            return

        return super().keypress(size, key)

class NoTabCheckBox(urwid.CheckBox):
    def keypress(self, size, key):
        if not isinstance(key, tuple) and key == 'tab':
            return 
        else:
            return super().keypress(size, key)

class FocusWidget(urwid.WidgetPlaceholder):

    def __init__(self, widget, max_focus_path_len, initial_focus, skip_focus):
        super().__init__(widget)
        self._initial_focus = tuple([ i for i in initial_focus ])
        self._skip_focus = tuple([ i for i in skip_focus ])
        self.max_focus_path_len = max_focus_path_len
    
    @property
    def skip_focus(self):
        return self._skip_focus
    
    @property
    def initial_focus(self):
        return list(self._initial_focus)
    
    @property
    def container(self):
        return self.original_widget.original_widget.original_widget

    def _can_set_focus_path(self, path):
        old = self.container.get_focus_path()
        try:
            self.container.set_focus_path(path)
            return True
        except IndexError:
            return False
        finally:
            self.container.set_focus_path(old)

    def _set_focus_path(self, path):
        self.container.set_focus_path(path)

    def _next_path(self, path: Iterable, ext=True) -> List:
        if len(path) < self.max_focus_path_len:
            next_path = [*path, 0]
            if self._can_set_focus_path(next_path):
                return next_path
        
        next_path = [*path[:-1], path[-1]+1]
        if self._can_set_focus_path(next_path):
            return next_path

        next_path = path
        while len(next_path) > 1 and next_path[-1] == 0:
            next_path = [*next_path[:-1]]
            if self._can_set_focus_path(next_path):
                return next_path
        if len(next_path) >= 2:
            next_path = [*next_path[:-2], next_path[-2]+1, 0]
            if self._can_set_focus_path(next_path):
                return next_path
            return self._next_path(next_path)
        if self._can_set_focus_path([*path[:-1], path[-1]+1]):
            return self._next_path(next_path)
        return

    def iter_focus_paths(self):
        path = None
        self._set_focus_path(self.initial_focus)
        while path != self.initial_focus:
            if path is None:
                path = self.container.get_focus_path()
            if path not in self.skip_focus:
                yield path
            path = self._next_path(path)
            if path is None:
                path = self._next_path([0])
            self._set_focus_path(path)
    
    def advance_focus(self, reverse=False):

        path = self.container.get_focus_path()

        paths = [ i for i in self.iter_focus_paths() ]

        if reverse:
            zipped_paths = zip(paths, [
                *paths[1:], paths[0]
            ])
        else:
            zipped_paths = zip([
                *paths[1:], paths[0]
            ], paths)
        
        next_path = map(lambda x: x[0], filter(
            lambda x: x[1] == path,
            zipped_paths
        ))
        p = next(next_path)
        if p in self.skip_focus:
            self.advance_focus(reverse=reverse)
            return
        self._set_focus_path(p)
        return

class AutoCompletePopUp(PopUpLauncher):
    _default_pop_up_parameters = {
        'left': 0,
        'top': 1,
        'overlay_width': 30,
        'overlay_height': 15,
    }
    def __init__(self,
        widget: Union[AutoCompleteEdit, AutoCompleteFloatEdit],
        apply_choice_cb: Callable[[str, str], None],
        close_cb: Callable[[], Any]
    ):
        super().__init__(widget)
        self._pop_up_parameters = dict(**self._default_pop_up_parameters)
        self.apply_choice_cb = apply_choice_cb
        self.close_cb = close_cb
        urwid.connect_signal(self._original_widget, 'open', lambda _, options: self._open_pop_up(options))

    def _open_pop_up(self, options):
        self.options = options
        max_height = self._default_pop_up_parameters['overlay_height']
        height = min(
            len(options),
            self._default_pop_up_parameters['overlay_height'],
        )
        if 0 < len(options) - height < max_height//3:
            height = 4 * max_height // 6
        
        self._pop_up_parameters = dict(**self._default_pop_up_parameters)
        self._pop_up_parameters.update({
            'overlay_width': max(
                self._original_widget._cache_maxcol,
                self._default_pop_up_parameters['overlay_width'],
            ),
            # account for header and footer rows
            'overlay_height': height+2,
        })
        self.open_pop_up()
    
    def _close_pop_up(self):
        self.close_pop_up()
        self.close_cb()

    def create_pop_up(self):
        pop_up = SuggestionPopup(
            self._original_widget.name, self.options,
            self.apply_choice_cb,
        )
        urwid.connect_signal(pop_up, 'close', lambda _: self._close_pop_up())
        return pop_up

    def get_pop_up_parameters(self):
        return self._pop_up_parameters

class SuggestionPopup(urwid.WidgetWrap):
    
    signals = ['close']

    def __init__(self,
        name: str,
        options: Iterable,
        apply_cb: Callable[[str, str], None]
    ):
        self.apply_cb = lambda _, v: apply_cb(name, v)
        body = []
        for c in options:
            button = urwid.Button(c)
            urwid.connect_signal(button, 'click', self.apply_cb, c)
            urwid.connect_signal(button, 'click', lambda _: self._emit("close"))
            body.append(urwid.AttrMap(button, None, focus_map='popup_focus'))
        walker = urwid.SimpleFocusListWalker(body, wrap_around=False)
        super().__init__(urwid.AttrWrap(IndicativeListBox(walker,
            topBar_endExposed_prop=("───", None, None),
            topBar_endCovered_prop=("▲ {} more ▲", None, None),
            bottomBar_endCovered_prop=("▼ {} more ▼", None, None),
            bottomBar_endExposed_prop=(f"─── {len(options)} ───", None, None),
        ), 'popup'))
            
        
    def keypress(self, size, key):
        if key == 'esc':
            self._emit("close")
            return

        if key == 'tab':
            return

        return super().keypress(size, key)
